@@ -51,8 +51,13 @@
 #include <openssl/err.h>
 
 #ifdef ENABLE_HTTP3
-#  include <ngtcp2/ngtcp2.h>
-#endif // ENABLE_HTTP3
+#  ifdef HAVE_LIBNGTCP2_CRYPTO_OPENSSL
+#    include <ngtcp2/ngtcp2_crypto_openssl.h>
+#  endif // HAVE_LIBNGTCP2_CRYPTO_OPENSSL
+#  ifdef HAVE_LIBNGTCP2_CRYPTO_BORINGSSL
+#    include <ngtcp2/ngtcp2_crypto_boringssl.h>
+#  endif // HAVE_LIBNGTCP2_CRYPTO_BORINGSSL
+#endif   // ENABLE_HTTP3
 
 #include "url-parser/url_parser.h"
 
@@ -128,7 +133,8 @@ Config::Config()
       unix_addr{},
       rps(0.),
       no_udp_gso(false),
-      max_udp_payload_size(0) {}
+      max_udp_payload_size(0),
+      ktls(false) {}
 
 Config::~Config() {
   if (addrs) {
@@ -338,11 +344,13 @@ void rps_cb(struct ev_loop *loop, ev_timer *w, int revents) {
     return;
   }
 
-  auto now = ev_now(loop);
+  auto now = std::chrono::steady_clock::now();
   auto d = now - client->rps_duration_started;
-  auto n = static_cast<size_t>(round(d * config.rps));
+  auto n = static_cast<size_t>(
+      round(std::chrono::duration<double>(d).count() * config.rps));
   client->rps_req_pending += n;
-  client->rps_duration_started = now - d + static_cast<double>(n) / config.rps;
+  client->rps_duration_started +=
+      util::duration_from(static_cast<double>(n) / config.rps);
 
   if (client->rps_req_pending == 0) {
     return;
@@ -419,10 +427,10 @@ void client_request_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
     return;
   }
 
-  ev_tstamp duration =
+  auto duration =
       config.timings[client->reqidx] - config.timings[client->reqidx - 1];
 
-  while (duration < 1e-9) {
+  while (duration < std::chrono::duration<double>(1e-9)) {
     if (client->submit_request() != 0) {
       ev_timer_stop(client->worker->loop, w);
       client->process_request_failure();
@@ -437,7 +445,7 @@ void client_request_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
         config.timings[client->reqidx] - config.timings[client->reqidx - 1];
   }
 
-  client->request_timeout_watcher.repeat = duration;
+  client->request_timeout_watcher.repeat = util::ev_tstamp_from(duration);
   ev_timer_again(client->worker->loop, &client->request_timeout_watcher);
 }
 } // namespace
@@ -464,7 +472,6 @@ Client::Client(uint32_t id, Worker *worker, size_t req_todo)
       local_addr{},
       new_connection_requested(false),
       final(false),
-      rps_duration_started(0),
       rps_req_pending(0),
       rps_req_inflight(0) {
   if (req_todo == 0) { // this means infinite number of requests are to be made
@@ -499,6 +506,8 @@ Client::Client(uint32_t id, Worker *worker, size_t req_todo)
   if (config.is_quic()) {
     quic.tx.data = std::make_unique<uint8_t[]>(64_k);
   }
+
+  ngtcp2_ccerr_default(&quic.last_error);
 #endif // ENABLE_HTTP3
 }
 
@@ -562,7 +571,6 @@ int Client::make_socket(addrinfo *addr) {
         ssl = SSL_new(worker->ssl_ctx);
       }
 
-      SSL_set_fd(ssl, fd);
       SSL_set_connect_state(ssl);
     }
   }
@@ -722,12 +730,17 @@ void Client::disconnect() {
   ev_io_stop(worker->loop, &wev);
   ev_io_stop(worker->loop, &rev);
   if (ssl) {
-    SSL_set_shutdown(ssl, SSL_get_shutdown(ssl) | SSL_RECEIVED_SHUTDOWN);
-    ERR_clear_error();
-
-    if (SSL_shutdown(ssl) != 1) {
+    if (config.is_quic()) {
       SSL_free(ssl);
       ssl = nullptr;
+    } else {
+      SSL_set_shutdown(ssl, SSL_get_shutdown(ssl) | SSL_RECEIVED_SHUTDOWN);
+      ERR_clear_error();
+
+      if (SSL_shutdown(ssl) != 1) {
+        SSL_free(ssl);
+        ssl = nullptr;
+      }
     }
   }
   if (fd != -1) {
@@ -1166,7 +1179,7 @@ int Client::connection_made() {
   if (config.rps_enabled()) {
     rps_watcher.repeat = std::max(0.01, 1. / config.rps);
     ev_timer_again(worker->loop, &rps_watcher);
-    rps_duration_started = ev_now(worker->loop);
+    rps_duration_started = std::chrono::steady_clock::now();
   }
 
   if (config.rps_enabled()) {
@@ -1190,9 +1203,9 @@ int Client::connection_made() {
     }
   } else {
 
-    ev_tstamp duration = config.timings[reqidx];
+    auto duration = config.timings[reqidx];
 
-    while (duration < 1e-9) {
+    while (duration < std::chrono::duration<double>(1e-9)) {
       if (submit_request() != 0) {
         process_request_failure();
         break;
@@ -1205,10 +1218,10 @@ int Client::connection_made() {
       }
     }
 
-    if (duration >= 1e-9) {
+    if (duration >= std::chrono::duration<double>(1e-9)) {
       // double check since we may have break due to reqidx wraps
       // around back to 0
-      request_timeout_watcher.repeat = duration;
+      request_timeout_watcher.repeat = util::ev_tstamp_from(duration);
       ev_timer_again(worker->loop, &request_timeout_watcher);
     }
   }
@@ -1308,6 +1321,8 @@ int Client::connected() {
   ev_io_stop(worker->loop, &wev);
 
   if (ssl) {
+    SSL_set_fd(ssl, fd);
+
     readfn = &Client::tls_handshake;
     writefn = &Client::tls_handshake;
 
@@ -1452,7 +1467,7 @@ int Client::write_udp(const sockaddr *addr, socklen_t addrlen,
 
   auto nwrite = sendmsg(fd, &msg, 0);
   if (nwrite < 0) {
-    if (nwrite == EAGAIN || nwrite == EWOULDBLOCK) {
+    if (errno == EAGAIN || errno == EWOULDBLOCK) {
       return 1;
     }
 
@@ -1957,9 +1972,10 @@ std::vector<std::string> read_uri_from_file(std::istream &infile) {
 } // namespace
 
 namespace {
-void read_script_from_file(std::istream &infile,
-                           std::vector<ev_tstamp> &timings,
-                           std::vector<std::string> &uris) {
+void read_script_from_file(
+    std::istream &infile,
+    std::vector<std::chrono::steady_clock::duration> &timings,
+    std::vector<std::string> &uris) {
   std::string script_line;
   int line_count = 0;
   while (std::getline(infile, script_line)) {
@@ -1992,7 +2008,9 @@ void read_script_from_file(std::istream &infile,
       exit(EXIT_FAILURE);
     }
 
-    timings.push_back(v / 1000.0);
+    timings.emplace_back(
+        std::chrono::duration_cast<std::chrono::steady_clock::duration>(
+            std::chrono::duration<double, std::milli>(v)));
     uris.push_back(script_line.substr(pos + 1, script_line.size()));
   }
 }
@@ -2278,6 +2296,7 @@ Options:
               Disable UDP GSO.
   --max-udp-payload-size=<SIZE>
               Specify the maximum outgoing UDP datagram payload size.
+  --ktls      Enable ktls.
   -v, --verbose
               Output debug information.
   --version   Display version information and exit.
@@ -2345,6 +2364,7 @@ int main(int argc, char **argv) {
         {"no-udp-gso", no_argument, &flag, 15},
         {"qlog-file-base", required_argument, &flag, 16},
         {"max-udp-payload-size", required_argument, &flag, 17},
+        {"ktls", no_argument, &flag, 18},
         {nullptr, 0, nullptr, 0}};
     int option_index = 0;
     auto c = getopt_long(argc, argv,
@@ -2354,44 +2374,65 @@ int main(int argc, char **argv) {
       break;
     }
     switch (c) {
-    case 'n':
-      config.nreqs = strtoul(optarg, nullptr, 10);
+    case 'n': {
+      auto n = util::parse_uint(optarg);
+      if (n == -1) {
+        std::cerr << "-n: bad option value: " << optarg << std::endl;
+        exit(EXIT_FAILURE);
+      }
+      config.nreqs = n;
       nreqs_set_manually = true;
       break;
-    case 'c':
-      config.nclients = strtoul(optarg, nullptr, 10);
+    }
+    case 'c': {
+      auto n = util::parse_uint(optarg);
+      if (n == -1) {
+        std::cerr << "-c: bad option value: " << optarg << std::endl;
+        exit(EXIT_FAILURE);
+      }
+      config.nclients = n;
       break;
+    }
     case 'd':
       datafile = optarg;
       break;
-    case 't':
+    case 't': {
 #ifdef NOTHREADS
       std::cerr << "-t: WARNING: Threading disabled at build time, "
                 << "no threads created." << std::endl;
 #else
-      config.nthreads = strtoul(optarg, nullptr, 10);
+      auto n = util::parse_uint(optarg);
+      if (n == -1) {
+        std::cerr << "-t: bad option value: " << optarg << std::endl;
+        exit(EXIT_FAILURE);
+      }
+      config.nthreads = n;
 #endif // NOTHREADS
       break;
-    case 'm':
-      config.max_concurrent_streams = strtoul(optarg, nullptr, 10);
+    }
+    case 'm': {
+      auto n = util::parse_uint(optarg);
+      if (n == -1) {
+        std::cerr << "-m: bad option value: " << optarg << std::endl;
+        exit(EXIT_FAILURE);
+      }
+      config.max_concurrent_streams = n;
       break;
+    }
     case 'w':
     case 'W': {
-      errno = 0;
-      char *endptr = nullptr;
-      auto n = strtoul(optarg, &endptr, 10);
-      if (errno == 0 && *endptr == '\0' && n < 31) {
-        if (c == 'w') {
-          config.window_bits = n;
-        } else {
-          config.connection_window_bits = n;
-        }
-      } else {
+      auto n = util::parse_uint(optarg);
+      if (n == -1 || n > 30) {
         std::cerr << "-" << static_cast<char>(c)
                   << ": specify the integer in the range [0, 30], inclusive"
                   << std::endl;
         exit(EXIT_FAILURE);
       }
+      if (c == 'w') {
+        config.window_bits = n;
+      } else {
+        config.connection_window_bits = n;
+      }
       break;
     }
     case 'f': {
@@ -2454,14 +2495,20 @@ int main(int argc, char **argv) {
       }
       break;
     }
-    case 'r':
-      config.rate = strtoul(optarg, nullptr, 10);
-      if (config.rate == 0) {
+    case 'r': {
+      auto n = util::parse_uint(optarg);
+      if (n == -1) {
+        std::cerr << "-r: bad option value: " << optarg << std::endl;
+        exit(EXIT_FAILURE);
+      }
+      if (n == 0) {
         std::cerr << "-r: the rate at which connections are made "
                   << "must be positive." << std::endl;
         exit(EXIT_FAILURE);
       }
+      config.rate = n;
       break;
+    }
     case 'T':
       config.conn_active_timeout = util::parse_duration_with_unit(optarg);
       if (!std::isfinite(config.conn_active_timeout)) {
@@ -2645,6 +2692,10 @@ int main(int argc, char **argv) {
         config.max_udp_payload_size = n;
         break;
       }
+      case 18:
+        // --ktls
+        config.ktls = true;
+        break;
       }
       break;
     default:
@@ -2856,15 +2907,33 @@ int main(int argc, char **argv) {
                   SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3 | SSL_OP_NO_COMPRESSION |
                   SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
 
+#ifdef SSL_OP_ENABLE_KTLS
+  if (config.ktls) {
+    ssl_opts |= SSL_OP_ENABLE_KTLS;
+  }
+#endif // SSL_OP_ENABLE_KTLS
+
   SSL_CTX_set_options(ssl_ctx, ssl_opts);
   SSL_CTX_set_mode(ssl_ctx, SSL_MODE_AUTO_RETRY);
   SSL_CTX_set_mode(ssl_ctx, SSL_MODE_RELEASE_BUFFERS);
 
   if (config.is_quic()) {
 #ifdef ENABLE_HTTP3
-    SSL_CTX_set_min_proto_version(ssl_ctx, TLS1_3_VERSION);
-    SSL_CTX_set_max_proto_version(ssl_ctx, TLS1_3_VERSION);
-#endif // ENABLE_HTTP3
+#  ifdef HAVE_LIBNGTCP2_CRYPTO_OPENSSL
+    if (ngtcp2_crypto_openssl_configure_client_context(ssl_ctx) != 0) {
+      std::cerr << "ngtcp2_crypto_openssl_configure_client_context failed"
+                << std::endl;
+      exit(EXIT_FAILURE);
+    }
+#  endif // HAVE_LIBNGTCP2_CRYPTO_OPENSSL
+#  ifdef HAVE_LIBNGTCP2_CRYPTO_BORINGSSL
+    if (ngtcp2_crypto_boringssl_configure_client_context(ssl_ctx) != 0) {
+      std::cerr << "ngtcp2_crypto_boringssl_configure_client_context failed"
+                << std::endl;
+      exit(EXIT_FAILURE);
+    }
+#  endif // HAVE_LIBNGTCP2_CRYPTO_BORINGSSL
+#endif   // ENABLE_HTTP3
   } else if (nghttp2::tls::ssl_ctx_set_proto_versions(
                  ssl_ctx, nghttp2::tls::NGHTTP2_TLS_MIN_VERSION,
                  nghttp2::tls::NGHTTP2_TLS_MAX_VERSION) != 0) {
