@@ -196,6 +196,28 @@ int Client::quic_extend_max_local_streams() {
   return 0;
 }
 
+namespace {
+int extend_max_stream_data(ngtcp2_conn *conn, int64_t stream_id,
+                           uint64_t max_data, void *user_data,
+                           void *stream_user_data) {
+  auto c = static_cast<Client *>(user_data);
+
+  if (c->quic_extend_max_stream_data(stream_id) != 0) {
+    return NGTCP2_ERR_CALLBACK_FAILURE;
+  }
+
+  return 0;
+}
+} // namespace
+
+int Client::quic_extend_max_stream_data(int64_t stream_id) {
+  auto s = static_cast<Http3Session *>(session.get());
+  if (s->unblock_stream(stream_id) != 0) {
+    return -1;
+  }
+  return 0;
+}
+
 namespace {
 int get_new_connection_id(ngtcp2_conn *conn, ngtcp2_cid *cid, uint8_t *token,
                           size_t cidlen, void *user_data) {
@@ -238,8 +260,10 @@ int generate_cid(ngtcp2_cid &dest) {
 } // namespace
 
 namespace {
-ngtcp2_tstamp timestamp(struct ev_loop *loop) {
-  return ev_now(loop) * NGTCP2_SECONDS;
+ngtcp2_tstamp quic_timestamp() {
+  return std::chrono::duration_cast<std::chrono::nanoseconds>(
+             std::chrono::steady_clock::now().time_since_epoch())
+      .count();
 }
 } // namespace
 
@@ -265,8 +289,9 @@ void rand(uint8_t *dest, size_t destlen, const ngtcp2_rand_ctx *rand_ctx) {
 } // namespace
 
 namespace {
-int recv_rx_key(ngtcp2_conn *conn, ngtcp2_crypto_level level, void *user_data) {
-  if (level != NGTCP2_CRYPTO_LEVEL_APPLICATION) {
+int recv_rx_key(ngtcp2_conn *conn, ngtcp2_encryption_level level,
+                void *user_data) {
+  if (level != NGTCP2_ENCRYPTION_LEVEL_1RTT) {
     return 0;
   }
 
@@ -338,7 +363,7 @@ int Client::quic_init(const sockaddr *local_addr, socklen_t local_addrlen,
       h2load::stream_reset,
       nullptr, // extend_max_remote_streams_bidi
       nullptr, // extend_max_remote_streams_uni
-      nullptr, // extend_max_stream_data
+      h2load::extend_max_stream_data,
       nullptr, // dcid_status
       nullptr, // handshake_confirmed
       nullptr, // recv_new_token
@@ -369,7 +394,7 @@ int Client::quic_init(const sockaddr *local_addr, socklen_t local_addrlen,
   if (config->verbose) {
     settings.log_printf = debug_log_printf;
   }
-  settings.initial_ts = timestamp(worker->loop);
+  settings.initial_ts = quic_timestamp();
   settings.rand_ctx.native_handle = &worker->randgen;
   if (!config->qlog_file_base.empty()) {
     assert(quic.qlog_file == nullptr);
@@ -384,7 +409,7 @@ int Client::quic_init(const sockaddr *local_addr, socklen_t local_addrlen,
       std::cerr << "Failed to open a qlog file: " << path << std::endl;
       return -1;
     }
-    settings.qlog.write = qlog_write_cb;
+    settings.qlog_write = qlog_write_cb;
   }
   if (config->max_udp_payload_size) {
     settings.max_tx_udp_payload_size = config->max_udp_payload_size;
@@ -453,7 +478,7 @@ void Client::quic_close_connection() {
 
   auto nwrite = ngtcp2_conn_write_connection_close(
       quic.conn, &ps.path, nullptr, buf.data(), buf.size(), &quic.last_error,
-      timestamp(worker->loop));
+      quic_timestamp());
 
   if (nwrite <= 0) {
     return;
@@ -463,7 +488,7 @@ void Client::quic_close_connection() {
             ps.path.remote.addrlen, buf.data(), nwrite, 0);
 }
 
-int Client::quic_write_client_handshake(ngtcp2_crypto_level level,
+int Client::quic_write_client_handshake(ngtcp2_encryption_level level,
                                         const uint8_t *data, size_t datalen) {
   int rv;
 
@@ -492,7 +517,7 @@ void quic_pkt_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
 
 int Client::quic_pkt_timeout() {
   int rv;
-  auto now = timestamp(worker->loop);
+  auto now = quic_timestamp();
 
   rv = ngtcp2_conn_handle_expiry(quic.conn, now);
   if (rv != 0) {
@@ -505,7 +530,7 @@ int Client::quic_pkt_timeout() {
 
 void Client::quic_restart_pkt_timer() {
   auto expiry = ngtcp2_conn_get_expiry(quic.conn);
-  auto now = timestamp(worker->loop);
+  auto now = quic_timestamp();
   auto t = expiry > now ? static_cast<ev_tstamp>(expiry - now) / NGTCP2_SECONDS
                         : 1e-9;
   quic.pkt_timer.repeat = t;
@@ -520,6 +545,8 @@ int Client::read_quic() {
   size_t pktcnt = 0;
   ngtcp2_pkt_info pi{};
 
+  auto ts = quic_timestamp();
+
   for (;;) {
     auto nread =
         recvfrom(fd, buf.data(), buf.size(), MSG_DONTWAIT, &su.sa, &addrlen);
@@ -542,8 +569,7 @@ int Client::read_quic() {
         },
     };
 
-    rv = ngtcp2_conn_read_pkt(quic.conn, &path, &pi, buf.data(), nread,
-                              timestamp(worker->loop));
+    rv = ngtcp2_conn_read_pkt(quic.conn, &path, &pi, buf.data(), nread, ts);
     if (rv != 0) {
       std::cerr << "ngtcp2_conn_read_pkt: " << ngtcp2_strerror(rv) << std::endl;
 
@@ -605,6 +631,7 @@ int Client::write_quic() {
   ngtcp2_path_storage_zero(&ps);
 
   auto s = static_cast<Http3Session *>(session.get());
+  auto ts = quic_timestamp();
 
   for (;;) {
     int64_t stream_id = -1;
@@ -629,8 +656,7 @@ int Client::write_quic() {
 
     auto nwrite = ngtcp2_conn_writev_stream(
         quic.conn, &ps.path, nullptr, bufpos, max_udp_payload_size, &ndatalen,
-        flags, stream_id, reinterpret_cast<const ngtcp2_vec *>(v), vcnt,
-        timestamp(worker->loop));
+        flags, stream_id, reinterpret_cast<const ngtcp2_vec *>(v), vcnt, ts);
     if (nwrite < 0) {
       switch (nwrite) {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
